/*
 ************************************************************************
 *
 * File 'tkCanvasPlus.c'.
 *
 * CanvasPlus widget implementation.
 *
 * Written by Jorge Lima in Nov 23 14:44:03 WET 2004
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include <X11/Xutil.h>
#include <tcl.h>
#include <tk.h>

#include "tkCanvasPlus.h"
#include "tkCanvasPlusDrawing.h"
#include "tkCanvasPlusTags.h"
#include "tkCanvasPlusModes.h"
#include "tkUtils.h"
#include "tkUnixDefault.h"
 
/*
 * Macros to test options
 */
#define OPT_NOP           0
#define OPT_REDRAW        BIT(1)
#define OPT_WIDTH         BIT(2)
#define OPT_HEIGHT        BIT(3)

/*
 * Information used for argv parsing.
 */

static Tk_OptionSpec configSpecs[] = {
  {TK_OPTION_PIXELS,  "-width",       "width",        "Width",
   DEF_CANVASP_WIDTH, -1, Tk_Offset(CanvasPlus, width),
   0, 0, OPT_REDRAW},
  {TK_OPTION_PIXELS,  "-height",      "height",       "Height",
   DEF_CANVASP_HEIGHT, -1, Tk_Offset(CanvasPlus, height),
   0, 0, OPT_REDRAW},
  {TK_OPTION_DOUBLE,  "-zoom",        "zoom",         "Zoom",
   DEF_CANVASP_ZOOM, -1, Tk_Offset(CanvasPlus, zoom),
   0, 0, OPT_REDRAW},
  {TK_OPTION_BOOLEAN, "-confine",     "confine",      "Confine",
   DEF_CANVASP_CONFINE, -1, Tk_Offset(CanvasPlus, confine),
   0, 0, OPT_NOP},
  {TK_OPTION_BOOLEAN, "-yinvert",     "yInvert",      "YInvert",
   DEF_CANVASP_Y_INVERT, -1, Tk_Offset(CanvasPlus, yInvert),
   0, 0, OPT_REDRAW},
  {TK_OPTION_STRING,  "-takefocus",   "takeFocus",    "TakeFocus",
   DEF_CANVASP_TAKE_FOCUS, -1, Tk_Offset(CanvasPlus, takeFocus),
   TK_CONFIG_NULL_OK, 0, OPT_NOP},
  {TK_OPTION_PIXELS,  "-closeenough", "closeEnough",  "CloseEnough",
   DEF_CANVASP_CLOSE_ENOUGH, -1, Tk_Offset(CanvasPlus, closeEnough),
   0, 0, OPT_NOP},
  {TK_OPTION_STRING, "-xscrollcommand", "xScrollCommand", "ScrollCommand",
   DEF_CANVASP_X_SCROLL_CMD, -1, Tk_Offset(CanvasPlus, xScrollCmd),
   TK_OPTION_NULL_OK, 0, OPT_NOP},
  {TK_OPTION_STRING, "-yscrollcommand", "yScrollCommand", "ScrollCommand",
   DEF_CANVASP_Y_SCROLL_CMD, -1, Tk_Offset(CanvasPlus, yScrollCmd),
   TK_OPTION_NULL_OK, 0, OPT_NOP},
  {TK_OPTION_SYNONYM, "-ce", NULL, NULL, NULL, 0, 0, 0,
   (ClientData)"-closeenough", 0},
  {TK_OPTION_END}
};

/*
 * Forward declarations for procedures defined later in this file:
 */

int	    CanvasPlusObjCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp,
					 int objc, Tcl_Obj * CONST objv[]));
static void CanvasPlusDeletedProc _ANSI_ARGS_((ClientData clientData));
static int  CanvasPlusConfigure _ANSI_ARGS_((Tcl_Interp *interp, CanvasPlus *cPtr,
					     int objc, Tcl_Obj * CONST objv[]));
static void CanvasPlusDestroy _ANSI_ARGS_((char *memPtr));
static void CanvasPlusDisplay _ANSI_ARGS_((ClientData clientData));
static void CanvasPlusEventProc _ANSI_ARGS_((ClientData clientData, XEvent *eventPtr));
static void CanvasPlusBindProc _ANSI_ARGS_((ClientData clientData, XEvent *eventPtr));
static int  CanvasPlusWidgetCmd _ANSI_ARGS_((ClientData clientData, Tcl_Interp *,
					    int objc, Tcl_Obj * CONST objv[]));
static Tcl_Obj *ScrollFractions _ANSI_ARGS_((int screen1, int screen2,
					     int object1, int object2));
static int CanvasPlusAdjustXview _ANSI_ARGS_((CanvasPlus *cPtr, int objc,
					     Tcl_Obj * CONST objv[]));
static int CanvasPlusAdjustYview _ANSI_ARGS_((CanvasPlus *cPtr, int objc,
					     Tcl_Obj * CONST objv[]));
static void CanvasPlusEventuallyRedraw _ANSI_ARGS_((CanvasPlus *cPtr,
						   int x1, int y1, int x2, int y2));
static void CanvasPlusSetOrigin _ANSI_ARGS_((CanvasPlus *cPtr, int xOrigin, int yOrigin));
static void CanvasPlusUpdateScrollbars _ANSI_ARGS_((CanvasPlus *canvasPtr));

static void PickCurrentItem _ANSI_ARGS_((CanvasPlus *canvas, XEvent *event));

static void ProcessEvent _ANSI_ARGS_((CanvasPlus *canvas, XEvent *event));


/*--------------------------------------------------------------
 * CanvasPlusCmd --
 *	This procedure is invoked to process the "canvas+" Tcl
 *	command.  It creates a new "canvas+" canvas widget.
 * Results:
 *	A standard Tcl result.
 * Side effects:
 *	A new widget is created and configured.
 *--------------------------------------------------------------
 */

int CanvasPlusObjCmd(clientData, interp, objc, objv)
     ClientData clientData;	/* Main window associated with
				 * interpreter. */
     Tcl_Interp *interp;	/* Current interpreter. */
     int objc;			/* Number of arguments. */
     Tcl_Obj * CONST objv[];	/* Argument objects. */
{
  CanvasPlus *cPtr;
  Tk_Window tkwin;
  Tk_OptionTable optionTable = (Tk_OptionTable) clientData;
  Tcl_CmdInfo info;
  char *commandName;
  
  if (optionTable == NULL) {
    optionTable = Tk_CreateOptionTable(interp, configSpecs);
    commandName = Tcl_GetStringFromObj(objv[0], (int *) NULL);
    Tcl_GetCommandInfo(interp, commandName, &info);
    info.clientData = (ClientData) optionTable;
    Tcl_SetCommandInfo(interp, commandName, &info);
  }
  
  if (objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "pathName ?options?");
    return TCL_ERROR;
  }

  tkwin = Tk_CreateWindowFromPath(interp, Tk_MainWindow(interp), 
				  Tcl_GetStringFromObj(objv[1], NULL), (char *) NULL);
  if (tkwin == NULL) {
    return TCL_ERROR;
  }
  
  /*
   * Allocate and initialize the widget record.
   */

  cPtr = (CanvasPlus *) ckalloc(sizeof(CanvasPlus));
  cPtr->widgetCmd = Tcl_CreateObjCommand(interp, 
					 Tk_PathName(tkwin), CanvasPlusWidgetCmd,
					 (ClientData) cPtr, CanvasPlusDeletedProc);
  cPtr->tkwin = tkwin;
  cPtr->interp = interp;
  cPtr->optionTable = optionTable;
  // Variables to manage view of canvas in screen window
  cPtr->xOrigin = 0;
  cPtr->yOrigin = 0;
  cPtr->xScrollCmd = NULL;
  cPtr->yScrollCmd = NULL;
  cPtr->Xmin = 0.0;
  cPtr->Xmax = 0.0;
  cPtr->Ymin = 0.0;
  cPtr->Ymax = 0.0;
  cPtr->scrollX1 = 0;
  cPtr->scrollX2 = 0;
  cPtr->scrollY1 = 0;
  cPtr->scrollY2 = 0;
  // Set mode  to a consistent state  
  cPtr->mode = NULL;
  // Scale factors for current screen.
  /* They will be initialized for true values later. */
  cPtr->x_scale = 1.0;
  cPtr->y_scale = 1.0;
  cPtr->l_scale = 1.0;
  /* Drawing tools */
  cPtr->copyGC = None;
  cPtr->drawable = None;
  cPtr->drawableMargin = 30;
  /* TO BE TESTED */
  cPtr->currentItem = NULL;
  cPtr->newCurrent = NULL;
  cPtr->closeEnough = 0;  // integer pixels instead of the original
  cPtr->pickEvent.type = LeaveNotify;
  cPtr->pickEvent.xcrossing.x = 0;
  cPtr->pickEvent.xcrossing.y = 0;
  cPtr->state = 0;
  cPtr->takeFocus = NULL;
  cPtr->flags = 0;
  //  cPtr->canvas_state = TK_STATE_NORMAL;
  /* END OF - TO BE TESTED */

  Tk_SetClass(tkwin, "CanvasPlus");
  /* Dont know if is realy necessary */
  //  Tk_SetClassProcs(canvasPtr->tkwin, &canvasClass, (ClientData) canvasPtr);
    
  Tk_CreateEventHandler(tkwin, ExposureMask|StructureNotifyMask|FocusChangeMask,
  			CanvasPlusEventProc, (ClientData) cPtr);
  Tk_CreateEventHandler(tkwin, KeyPressMask|KeyReleaseMask
  			|ButtonPressMask|ButtonReleaseMask|EnterWindowMask
  			|LeaveWindowMask|PointerMotionMask|VirtualEventMask,
  			CanvasPlusBindProc, (ClientData) cPtr);
  //  Tk_CreateSelHandler(tkwin, XA_PRIMARY, XA_STRING, CanvasFetchSelection,
  //	      (ClientData) canvasPtr, XA_STRING);

  /* Init option table */
  if (Tk_InitOptions(interp, (char *) cPtr, optionTable, tkwin) != TCL_OK) {
    Tk_DestroyWindow(cPtr->tkwin);
    ckfree((char *) cPtr);
    return TCL_ERROR;
  }

  /* Set options supplied in the command line
   * perform post configurations. */
  if (CanvasPlusConfigure(interp, cPtr, objc - 2, objv + 2) != TCL_OK) {
    goto error;
  }
 
  Tcl_SetObjResult(interp, Tcl_NewStringObj(Tk_PathName(cPtr->tkwin),-1));
  return TCL_OK;
  
 error:
  Tk_DestroyWindow(cPtr->tkwin);
  return TCL_ERROR;
}


/*--------------------------------------------------------------
 * CanvasPlusWidgetCmd --
 *	This procedure is invoked to process the Tcl command
 *	that corresponds to a widget managed by this module.
 *	See the user documentation for details on what it does.
 * Results:
 *	A standard Tcl result.
 * Side effects:
 *	See the user documentation.
 *-------------------------------------------------------------*/
static int CanvasPlusWidgetCmd(clientData, interp, objc, objv)
     ClientData clientData;		/* Information about square widget. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int objc;				/* Number of arguments. */
     Tcl_Obj * CONST objv[];		/* Argument objects. */
{
  CanvasPlus *cPtr = (CanvasPlus *) clientData;
  int result = TCL_OK;
  const char *options[] = {
    "bind",  "cget",  "clearmode",  "configure",  "get",  "modes",  "pan",
    "redraw",  "setmode",  "xview",  "yview",  "zoom", (char *) NULL
  };
  enum {
    CP_BIND, CP_CGET, CP_CLRMODE, CP_CONFIGURE, CP_GET, CP_MODES, CP_PAN,
    CP_REDRAW, CP_SETMODE, CP_XVIEW, CP_YVIEW, CP_ZOOM
  };

  Tcl_Obj *resultObjPtr;
  int index;

  if (objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "option ?arg arg...?");
    return TCL_ERROR;
  }
  if (Tcl_GetIndexFromObj(interp,objv[1],options,"command",0,&index)!=TCL_OK) {
    return TCL_ERROR;
  }

  PROTECT(cPtr); // avoid memory deletion while in use

  switch (index) {
  case CP_BIND:
    {
      ClientData object = 0;
      CP_Mode *mode = cPtr->mode;
      if (mode == NULL) {
	Tcl_AppendResult(interp, "no mode for bindings", (char *) NULL);	
	goto error;
      }
      if ((objc < 3) || (objc > 5)) {
	Tcl_WrongNumArgs(interp, 2, objv, "tagOrId ?sequence? ?command?");
	goto error;
      }
      
      // Figure out what object to use for the binding (individual item vs. tag).      
      object = CPGetTagOrId(cPtr->mode, Tcl_GetString(objv[2]));
      if (object == 0) {
	Tcl_AppendResult(interp, "item \"", Tcl_GetString(objv[2]),
			 "\" doesn't exist", (char *) NULL);
	goto error;
      }
      
      if (objc == 5) {
	char* sequence = Tcl_GetStringFromObj(objv[3],NULL);
	char* script = Tcl_GetStringFromObj(objv[4],NULL);
	result = CPBindScriptToTag(interp, mode->bindingTable,
				   object, sequence, script);
	goto done;
      } else if (objc == 4) {
	char* sequence = Tcl_GetStringFromObj(objv[3],NULL);
	result = CPGetBindScript(interp, mode->bindingTable,
				 object, sequence);
	goto done;
      } else {
	Tk_GetAllBindings(interp, mode->bindingTable, object);
      }
    }
    break;
  case CP_CGET:
    if (objc != 3) {
      Tcl_WrongNumArgs(interp, 2, objv, "option");
      goto error;
    }
    resultObjPtr =
      Tk_GetOptionValue(interp, (char *) cPtr,
			cPtr->optionTable, objv[2], cPtr->tkwin);
    if (resultObjPtr == NULL) goto error;
    Tcl_SetObjResult(interp, resultObjPtr);
    break;
  case CP_CLRMODE:
    CPUnsetMode(cPtr);
    cPtr->currentItem = NULL;
    cPtr->newCurrent = NULL;
    break;
  case CP_CONFIGURE:
    resultObjPtr = NULL;
    if (objc == 2) {
      resultObjPtr = Tk_GetOptionInfo(interp, (char *) cPtr,
				      cPtr->optionTable, (Tcl_Obj *) NULL,
				      cPtr->tkwin);
      if (resultObjPtr == NULL) goto error;
    } else if (objc == 3) {
      resultObjPtr = Tk_GetOptionInfo(interp, (char *) cPtr,
				      cPtr->optionTable, objv[2], cPtr->tkwin);
      if (resultObjPtr == NULL) goto error;
    } else {
      result = CanvasPlusConfigure(interp, cPtr, objc - 2, objv + 2);
    }
    if (resultObjPtr != NULL) {
      Tcl_SetObjResult(interp, resultObjPtr);
    }
    break;
  case CP_GET:
    {
      CP_Mode *mode = cPtr->mode;
      if (mode == NULL) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("mode is undefined",-1));
	goto error;
      }
      if (objc == 2) {
	Tcl_SetObjResult(interp, CPListAllItems(mode));
      }
      else if (objc == 3) {
	Tk_Uid tagUid = CPGetTagUid(Tcl_GetString(objv[2]));
	Tcl_SetObjResult(interp, CPListTaggedItems(mode, tagUid));
      }
      else {
	Tcl_WrongNumArgs(interp, 2, objv, "?tag?");
	goto error;
      }
    }
    break;
  case CP_MODES:
    Tcl_SetObjResult(interp, CPListModes());
    break;
  case CP_PAN:
    {
      int x, y;
      if (objc != 4) {
	Tcl_WrongNumArgs(interp, 2, objv, "win_x win_y");
	goto error;
      }
      if(Tcl_GetIntFromObj(interp, objv[2], &x)==TCL_ERROR) {
	goto error;
      }
      if(Tcl_GetIntFromObj(interp, objv[3], &y)==TCL_ERROR) {
	goto error;
      }
      x -= Tk_Width(cPtr->tkwin)/2;
      y -= Tk_Height(cPtr->tkwin)/2;
      CanvasPlusSetOrigin(cPtr, cPtr->xOrigin + x, cPtr->yOrigin + y);
      cPtr->flags |= UPDATE_SCROLLBARS;
      CanvasPlusEventuallyRedraw(cPtr, 0, 0, Tk_Width(cPtr->tkwin),
				 Tk_Height(cPtr->tkwin));
    }
    break;
  case CP_REDRAW:
    cPtr->flags |= UPDATE_SCROLLBARS;
    CanvasPlusEventuallyRedraw(cPtr, 0, 0, Tk_Width(cPtr->tkwin),
			       Tk_Height(cPtr->tkwin));
    break;
  case CP_SETMODE:
    if (objc==3) {
      Tcl_Obj *iname = CPGetUniqCmdName(interp, "c+_interp");
      if(CPSetMode(cPtr, iname, objv[2])==TCL_ERROR) goto error;
      Tcl_SetObjResult(interp, iname);
      cPtr->currentItem = NULL;
      cPtr->newCurrent = NULL;
    }
    else if (objc==4) {
      if(CPSetMode(cPtr, objv[3], objv[2])==TCL_ERROR) goto error;
      Tcl_SetObjResult(interp, objv[3]);
    }
    else {
      Tcl_WrongNumArgs(interp, 2, objv, "mode_name");
      goto error;
    }
    break;
  case CP_XVIEW:
    if (CanvasPlusAdjustXview(cPtr, objc, objv)==TCL_ERROR) goto error;
    break;
  case CP_YVIEW:
    if (CanvasPlusAdjustYview(cPtr, objc, objv)==TCL_ERROR) goto error;
    break;
  case CP_ZOOM:
    if (objc != 3) {
      Tcl_WrongNumArgs(interp, 2, objv, "zoom_factor");
      goto error;
    }
    if(Tcl_GetDoubleFromObj(interp, objv[2], &cPtr->zoom)==TCL_ERROR) {
      goto error;
    }
    CPRecomputeCanvasDimensions(cPtr);
    cPtr->flags |= UPDATE_SCROLLBARS;
    CanvasPlusEventuallyRedraw(cPtr, 0, 0, Tk_Width(cPtr->tkwin),
			       Tk_Height(cPtr->tkwin));
    break;
  }
 done:
  UNPROTECT(cPtr);
  return result;

 error:
  UNPROTECT(cPtr);
  return TCL_ERROR;
}

/*----------------------------------------------------------------------
 * CanvasPlusAdjustXview --
 *	This procedure will draw the canvas+ contents
 * Results:
 *	None.
 * Side effects:
 *	Everything in the visible region of canvas+ will be redrawn.
 *---------------------------------------------------------------------*/
static int CanvasPlusAdjustXview(CanvasPlus *cPtr, int objc, Tcl_Obj * CONST objv[])
{
  int count, type;
  int newX = 0;
  double fraction;

  if (objc == 2) {
    Tcl_SetObjResult(cPtr->interp, 
		     ScrollFractions(cPtr->xOrigin,
				     cPtr->xOrigin + Tk_Width(cPtr->tkwin),
				     cPtr->scrollX1,
				     cPtr->scrollX2));
  } else {
    CONST char **args = GetStringsFromObjs(objc, objv);
    type = Tk_GetScrollInfo(cPtr->interp, objc, args, &fraction, &count);
    if (args) ckfree((char *) args);
    switch (type) {
    case TK_SCROLL_ERROR:
      return TCL_ERROR;
    case TK_SCROLL_MOVETO:
      newX = cPtr->scrollX1 + (int)(fraction*(cPtr->scrollX2-cPtr->scrollX1)+0.5);
      break;
    case TK_SCROLL_PAGES:
      newX = (int)(cPtr->xOrigin + count * .9 * (Tk_Width(cPtr->tkwin)));
      break;
    case TK_SCROLL_UNITS:
      newX = (int)(cPtr->xOrigin + count * .1*(Tk_Width(cPtr->tkwin)));
      break;
    }
    CanvasPlusSetOrigin(cPtr, newX, cPtr->yOrigin);

  }
  return TCL_OK;
}

/*----------------------------------------------------------------------
 * CanvasPlusAdjustYview --
 *	This procedure will draw the canvas+ contents
 * Results:
 *	None.
 * Side effects:
 *	Everything in the visible region of canvas+ will be redrawn.
 *---------------------------------------------------------------------*/
static int CanvasPlusAdjustYview(CanvasPlus *cPtr, int objc, Tcl_Obj * CONST objv[])
{
  int count, type;
  int newY = 0;
  double fraction;

  if (objc == 2) {
    Tcl_SetObjResult(cPtr->interp, 
		     ScrollFractions(cPtr->yOrigin,
				     cPtr->yOrigin + Tk_Height(cPtr->tkwin),
				     cPtr->scrollY1,
				     cPtr->scrollY2));
  } else {
    CONST char **args = GetStringsFromObjs(objc, objv);
    type = Tk_GetScrollInfo(cPtr->interp, objc, args, &fraction, &count);
    if (args) ckfree((char *) args);
    switch (type) {
    case TK_SCROLL_ERROR:
      return TCL_ERROR;
    case TK_SCROLL_MOVETO:
      newY = cPtr->scrollY1 + (int)(fraction*(cPtr->scrollY2-cPtr->scrollY1)+0.5);
      break;
    case TK_SCROLL_PAGES:
      newY = (int)(cPtr->yOrigin + count * .9 * (Tk_Height(cPtr->tkwin)));
      break;
    case TK_SCROLL_UNITS:
      newY = (int)(cPtr->yOrigin + count * .1*(Tk_Height(cPtr->tkwin)));
      break;
    }
    CanvasPlusSetOrigin(cPtr, cPtr->xOrigin, newY);
  }
  return TCL_OK;
}


/*----------------------------------------------------------------------
 * CanvasPlusConfigure --
 *	This procedure is called to perform post config operations
 * Results:
 *	The return value is a standard Tcl result.  If TCL_ERROR is
 *	returned, then the interp's result contains an error message.
 * Side effects:
 *	Configuration information, such as colors, border width,
 *	etc. get set for cPtr;
 *---------------------------------------------------------------------*/
static int CanvasPlusConfigure(Tcl_Interp *interp, CanvasPlus *cPtr,
			       int objc, Tcl_Obj * CONST objv[])
{
  Tk_Window tkwin = cPtr->tkwin;
  if (Tk_SetOptions(interp, (char *) cPtr, cPtr->optionTable, objc,
		    objv, tkwin, NULL, &cPtr->optMask) != TCL_OK) {
    return TCL_ERROR;
  }
  /* make sure the cp->y_scale has the correct sign */
  if (cPtr->yInvert)
    cPtr->y_scale = -fabs(cPtr->y_scale);
  else
    cPtr->y_scale = fabs(cPtr->y_scale);

  /* if zoom has changed do it (check if it has changed) */
  if (cPtr->optMask & OPT_REDRAW) {
    CPRecomputeCanvasDimensions(cPtr);
    /* Set the background for the window. */
    Tk_SetWindowBackground(tkwin, 0);
    Tk_GeometryRequest(tkwin, cPtr->width, cPtr->height);
    /* Then arrange for the window to be redisplayed. */
    CanvasPlusSetOrigin(cPtr, cPtr->xOrigin, cPtr->yOrigin);
    /* Previous command might have been a NOP. So make sure the redraw is scheduled */
    cPtr->flags |= UPDATE_SCROLLBARS;
    CanvasPlusEventuallyRedraw(cPtr, 0, 0, Tk_Width(tkwin), Tk_Height(tkwin));
  }
  return TCL_OK;
}


/*--------------------------------------------------------------
 * CanvasPlusEventProc --
 *	This procedure is invoked by the Tk dispatcher for
 *	various events on canvas+.
 * Results:
 *	None.
 * Side effects:
 *	When the window gets deleted, internal structures get
 *	cleaned up.  When it gets exposed, it is redisplayed.
 *--------------------------------------------------------------*/
static void CanvasPlusEventProc(clientData, eventPtr)
    ClientData clientData;	/* Information about window. */
    XEvent *eventPtr;		/* Information about event. */
{
  CanvasPlus *cPtr = (CanvasPlus *) clientData;

  if (eventPtr->type == Expose) {
    int x1, x2, y1, y2;
    x1 = eventPtr->xexpose.x;
    y1 = eventPtr->xexpose.y;
    x2 = x1 + eventPtr->xexpose.width;
    y2 = y1 + eventPtr->xexpose.height;
    CanvasPlusEventuallyRedraw(cPtr, x1, y1, x2, y2);
  } else if (eventPtr->type == ConfigureNotify) {
    int w, h;
    w = Tk_Width(cPtr->tkwin);
    h = Tk_Height(cPtr->tkwin);
    cPtr->flags |= UPDATE_SCROLLBARS;
    CanvasPlusEventuallyRedraw(cPtr, 0, 0, w, h);
  } else if (eventPtr->type == DestroyNotify) {
    if (cPtr->tkwin != NULL) {
      if (cPtr->mode != NULL)
	CPUnsetMode(cPtr);
      Tk_FreeConfigOptions((char *) cPtr, cPtr->optionTable, cPtr->tkwin);
      if (cPtr->copyGC != None)
      	Tk_FreeGC(Tk_Display(cPtr->tkwin), cPtr->copyGC);
      cPtr->tkwin = NULL;
      Tcl_DeleteCommandFromToken(cPtr->interp, cPtr->widgetCmd);
    }
    if (cPtr->flags & REDRAW_PENDING) {
      Tcl_CancelIdleCall(CanvasPlusDisplay, (ClientData) cPtr);
    }
    Tcl_EventuallyFree((ClientData) cPtr, CanvasPlusDestroy);
  }
}

/*--------------------------------------------------------------
 * CanvasPlusEventuallyRedraw --
 *
 *	Arrange for part or all of a canvas+ widget to redrawn at
 *	some convenient time in the future.
 * Results:
 *	None.
 * Side effects:
 *	The screen will eventually be refreshed.
 *-------------------------------------------------------------*/
static void CanvasPlusEventuallyRedraw(CanvasPlus *cPtr, int x1, int y1, int x2, int y2)
{
  if (cPtr->tkwin == NULL) return;

  if ((x1 >= x2) || (y1 >= y2) || (x2 < 0) || (y2 < 0) ||
      (x1 >= Tk_Width(cPtr->tkwin)) || (y1 >= Tk_Height(cPtr->tkwin))) {
    return;
  }
  if (cPtr->flags & BBOX_NOT_EMPTY) {
    if (x1 < cPtr->redrawX1) cPtr->redrawX1 = x1;
    if (y1 < cPtr->redrawY1) cPtr->redrawY1 = y1;
    if (x2 > cPtr->redrawX2) cPtr->redrawX2 = x2;
    if (y2 > cPtr->redrawY2) cPtr->redrawY2 = y2;
  } else {
    cPtr->redrawX1 = x1;
    cPtr->redrawY1 = y1;
    cPtr->redrawX2 = x2;
    cPtr->redrawY2 = y2;
    cPtr->flags |= BBOX_NOT_EMPTY;
  }
  if (!(cPtr->flags & REDRAW_PENDING)) {
    Tcl_DoWhenIdle(CanvasPlusDisplay, (ClientData) cPtr);
    cPtr->flags |= REDRAW_PENDING;
  }
}

void CPForceRedraw(CanvasPlus *cPtr)
{
  Tk_Window tkwin = cPtr->tkwin;
  cPtr->flags |= UPDATE_SCROLLBARS;
  CanvasPlusEventuallyRedraw(cPtr, 0, 0, Tk_Width(tkwin), Tk_Height(tkwin));
}

/*--------------------------------------------------------------
 * CanvasPlusDisplay --
 *	This procedure redraws the contents of a canvas+ window.
 *	It is invoked as a do-when-idle handler, so it only runs
 *	when there's nothing else for the application to do.
 * Results:
 *	None.
 * Side effects:
 *	Information appears on the screen.
 *--------------------------------------------------------------*/
static void CanvasPlusDisplay(ClientData clientData)
{
  int x0, y0, width, height, margin;
  CanvasPlus *cPtr = (CanvasPlus *) clientData;
  Tk_Window tkwin = cPtr->tkwin;

  if (cPtr->tkwin == NULL) return;

  PROTECT(cPtr);

  if (Tk_IsMapped(tkwin)) {
    // Create a graphics context for copy PIXMAP operations.
    if (cPtr->copyGC == None) {
      //XColor *color = Tk_GetColor(cPtr->interp, tkwin, "red");
      XGCValues gcValues;
      gcValues.function = GXcopy;
      gcValues.graphics_exposures = False;
      gcValues.background = 0;
      gcValues.foreground = 0;
      cPtr->copyGC = Tk_GetGC(tkwin,GCFunction|GCGraphicsExposures|
			      GCBackground|GCForeground, &gcValues);
    }
    
    /*
     * Compute the intersection between the area that needs redrawing
     * and the area that's visible on the screen.
     */
    // safety margin around the redraw area in the drawable
    margin = cPtr->drawableMargin;

    x0 = cPtr->redrawX1;
    y0 = cPtr->redrawY1;
    
    if (cPtr->redrawX2 > Tk_Width(tkwin)) cPtr->redrawX2 = Tk_Width(tkwin);
    if (cPtr->redrawY2 > Tk_Height(tkwin)) cPtr->redrawY2 = Tk_Height(tkwin);
    
    // Coordinate conversion rotines need to know where we are drawing
    cPtr->drawableXOrigin = cPtr->xOrigin + x0 - margin;
    cPtr->drawableYOrigin = cPtr->yOrigin + y0 - margin;
    
    width = cPtr->redrawX2 - cPtr->redrawX1;
    height = cPtr->redrawY2 - cPtr->redrawY1;

    cPtr->drawable = Tk_GetPixmap(Tk_Display(tkwin), Tk_WindowId(tkwin),
				  width + 2*margin, height + 2*margin,
				  Tk_Depth(tkwin));
    
    XFillRectangle(Tk_Display(tkwin), cPtr->drawable, cPtr->copyGC,
		   margin, margin, (unsigned int) width, (unsigned int) height);  
    
    CPDrawContents(cPtr);
    
    XCopyArea(Tk_Display(tkwin), cPtr->drawable, Tk_WindowId(tkwin),
	      cPtr->copyGC, margin, margin, (unsigned int) width,
	      (unsigned int) height, x0, y0);

    Tk_FreePixmap(Tk_Display(tkwin), cPtr->drawable);
    cPtr->drawable = None;
  }

  cPtr->flags &= ~(REDRAW_PENDING|BBOX_NOT_EMPTY);
  cPtr->redrawX1 = cPtr->redrawX2 = 0;
  cPtr->redrawY1 = cPtr->redrawY2 = 0;
  if (cPtr->flags & UPDATE_SCROLLBARS) {
    CanvasPlusUpdateScrollbars(cPtr);
  }

  UNPROTECT(cPtr);
}

/*-------------------------------------------------------------- 
 * CanvasUpdateScrollbars --
 *	This procedure is invoked whenever a canvas+ has changed in
 *	a way that requires scrollbars to be redisplayed (e.g. the
 *	view in the canvas has changed).
 * Results:
 *	None.
 * Side effects:
 *	If there are scrollbars associated with the canvas+, then
 *	their scrolling commands are invoked to cause them to
 *	redisplay.  If errors occur, additional Tcl commands may
 *	be invoked to process the errors.
 *-------------------------------------------------------------*/
static void CanvasPlusUpdateScrollbars(CanvasPlus *canvasPtr)
{
    Tcl_Interp *interp;
    int res, xOrigin, yOrigin, width, height;
    int scrollX1, scrollX2, scrollY1, scrollY2;
    char *xScrollCmd, *yScrollCmd;
    /*
     * Save all the relevant values from the canvasPtr, because it might be
     * deleted as part of either of the two calls to Tcl_VarEval below.
     */ 
    interp = canvasPtr->interp;
    PROTECT(interp);
    xScrollCmd = canvasPtr->xScrollCmd;
    if (xScrollCmd != (char *) NULL) PROTECT(xScrollCmd);
    yScrollCmd = canvasPtr->yScrollCmd;
    if (yScrollCmd != (char *) NULL) PROTECT(yScrollCmd);
    xOrigin = canvasPtr->xOrigin;
    yOrigin = canvasPtr->yOrigin;
    width = Tk_Width(canvasPtr->tkwin);
    height = Tk_Height(canvasPtr->tkwin);
    scrollX1 = canvasPtr->scrollX1;
    scrollX2 = canvasPtr->scrollX2;
    scrollY1 = canvasPtr->scrollY1;
    scrollY2 = canvasPtr->scrollY2;
    canvasPtr->flags &= ~UPDATE_SCROLLBARS;
    if (xScrollCmd != NULL) {
	Tcl_Obj *frac = ScrollFractions(xOrigin, xOrigin+width, scrollX1, scrollX2);
	res = Tcl_VarEval(interp, xScrollCmd, " ", Tcl_GetString(frac), (char *)NULL);
	Tcl_DecrRefCount(frac);
	if (res != TCL_OK) Tcl_BackgroundError(interp);
	Tcl_ResetResult(interp);
        UNPROTECT(xScrollCmd);
    }

    if (yScrollCmd != NULL) {
	Tcl_Obj *frac = ScrollFractions(yOrigin, yOrigin+height, scrollY1, scrollY2);
	res = Tcl_VarEval(interp, yScrollCmd, " ", Tcl_GetString(frac), (char *)NULL);
	Tcl_DecrRefCount(frac);
	if (res != TCL_OK) Tcl_BackgroundError(interp);
	Tcl_ResetResult(interp);
        UNPROTECT(yScrollCmd);
    }
    UNPROTECT(interp);
}


/*--------------------------------------------------------------
 * CanvasPlusSetOrigin --
 *	This procedure is invoked to change the mapping between
 *	canvas coordinates and screen coordinates in the canvas
 *	window.
 * Results:
 *	None.
 * Side effects:
 *	The canvas will be redisplayed to reflect the change in
 *	view.  In addition, scrollbars will be updated if there
 *	are any.
 *-------------------------------------------------------------*/
static void CanvasPlusSetOrigin(CanvasPlus *cPtr, int xOrigin, int yOrigin)
{
  int left, right, top, bottom, delta;

  if (cPtr->confine) {
    left = xOrigin - cPtr->scrollX1;
    right = cPtr->scrollX2 - (xOrigin + Tk_Width(cPtr->tkwin));
    top = yOrigin - cPtr->scrollY1;
    bottom = cPtr->scrollY2 - (yOrigin + Tk_Height(cPtr->tkwin));
    if ((left < 0) && (right > 0)) {
      delta = (right > -left) ? -left : right;
      xOrigin += delta;
    } else if ((right < 0) && (left > 0)) {
      delta = (left > -right) ? -right : left;
      xOrigin -= delta;
    }
    if ((top < 0) && (bottom > 0)) {
      delta = (bottom > -top) ? -top : bottom;
      yOrigin += delta;
    } else if ((bottom < 0) && (top > 0)) {
      delta = (top > -bottom) ? -bottom : top;
      yOrigin -= delta;
    }
  }
  if ((xOrigin == cPtr->xOrigin) && (yOrigin == cPtr->yOrigin))
    return;

  CanvasPlusEventuallyRedraw(cPtr, 0, 0, Tk_Width(cPtr->tkwin),
			     Tk_Height(cPtr->tkwin));
  cPtr->xOrigin = xOrigin;
  cPtr->yOrigin = yOrigin;
  cPtr->flags |= UPDATE_SCROLLBARS;
  CanvasPlusEventuallyRedraw(cPtr, 0, 0, Tk_Width(cPtr->tkwin),
			     Tk_Height(cPtr->tkwin));
}

/*----------------------------------------------------------------------
 * CanvasPlusDestroy --
 *	This procedure is invoked by Tcl_EventuallyFree or Tcl_Release
 *	to clean up the internal structure of a canvas+ at a safe time
 *	(when no-one is using it anymore).
 * Results:
 *	None.
 * Side effects:
 *	Everything associated with the canvas+ is freed up.
 *---------------------------------------------------------------------*/
static void CanvasPlusDestroy(char *memPtr) /* Info about. */
{
  CanvasPlus *cPtr = (CanvasPlus *) memPtr;
  Tk_DeleteOptionTable(cPtr->optionTable);
  ckfree((char *) cPtr);
}

/*----------------------------------------------------------------------
 * CanvasPlusDeletedProc --
 *	This procedure is invoked when a widget command is deleted.  If
 *	the widget isn't already in the process of being destroyed,
 *	this command destroys it.
 * Results:
 *	None.
 * Side effects:
 *	The widget is destroyed.
 *----------------------------------------------------------------------*/
static void CanvasPlusDeletedProc(clientData)
    ClientData clientData;	/* Pointer to widget record for widget. */
{
    CanvasPlus *cPtr = (CanvasPlus *) clientData;
    Tk_Window tkwin = cPtr->tkwin;
    /*
     * This procedure could be invoked either because the window was
     * destroyed and the command was then deleted (in which case tkwin
     * is NULL) or because the command was deleted, and then this procedure
     * destroys the widget.
     */
    if (tkwin != NULL)
	Tk_DestroyWindow(tkwin);
}


/*
 * Here we don't use the CXI argument processing approach to be able to
 * use the TK builtint option processing faculities.
 */

int Canvasplus_Init(Tcl_Interp *interp)
{
  Tcl_CreateObjCommand(interp, "canvas+",
		       CanvasPlusObjCmd, (ClientData)NULL,
		       (Tcl_CmdDeleteProc *) NULL);
  return TCL_OK;
}

/*******************************************************************************/
/** BINDINGS END EVENT PROCESSING ** TO BE TESTED ** VERY BETA ** BASED IN TK **/
/*******************************************************************************/

/*--------------------------------------------------------------
 * CanvasBindProc --
 *	This procedure is invoked by the Tk dispatcher to handle
 *	events associated with bindings on items.
 * Results:
 *	None.
 * Side effects:
 *	Depends on the command invoked as part of the binding
 *	(if there was any).
 *--------------------------------------------------------------
 */
static void CanvasPlusBindProc(ClientData clientData, XEvent *event)
{
  CanvasPlus *canvas = (CanvasPlus *) clientData;

  PROTECT(canvas);
  /* This code below keeps track of the current modifier state in
   * canvasPtr>state.  This information is used to defer repicks of
   * the current item while buttons are down.
   */
  if ((event->type == ButtonPress) || (event->type == ButtonRelease)) {
      int mask;
      switch (event->xbutton.button) {
      case Button1: mask = Button1Mask; break;
      case Button2: mask = Button2Mask; break;
      case Button3: mask = Button3Mask; break;
      case Button4: mask = Button4Mask; break;
      case Button5: mask = Button5Mask; break;
      default: mask = 0;
      }
      if (event->type == ButtonPress) {
	/* On a button press, first repick the current item using
	 * the button state before the event, then process the event.*/  
	canvas->state = event->xbutton.state;
	PickCurrentItem(canvas, event);
	canvas->state ^= mask;
	ProcessEvent(canvas, event);
      } else {
	/* Button release: first process the event, with the button
	 * still considered to be down.  Then repick the current
	 * item under the assumption that the button is no longer down. */      
	canvas->state = event->xbutton.state;
	ProcessEvent(canvas, event);
	event->xbutton.state ^= mask;
	canvas->state = event->xbutton.state;
	PickCurrentItem(canvas, event);
	event->xbutton.state ^= mask;
      }
  } else if ((event->type == EnterNotify) || (event->type == LeaveNotify)) {
    canvas->state = event->xcrossing.state;
    PickCurrentItem(canvas, event);
  } else if (event->type == MotionNotify) {
    canvas->state = event->xmotion.state;
    PickCurrentItem(canvas, event);
    ProcessEvent(canvas, event);
  } else {
    ProcessEvent(canvas, event);
  }

  UNPROTECT(canvas);
}

/*--------------------------------------------------------------
 * PickCurrentItem --
 *	Find the topmost item in a canvas that contains a given
 *	location and mark the the current item.  If the current
 *	item has changed, generate a fake exit event on the old
 *	current item, a fake enter event on the new current item
 *	item and force a redraw of the two items. Canvas items
 *      that are hidden or disabled are ignored.
 * Results:
 *	None.
 * Side effects:
 *	The current item for canvas may change.  If it does,
 *	then the commands associated with item entry and exit
 *	could do just about anything.  A binding script could
 *	delete the canvas, so callers should protect themselves
 *	with Tcl_Preserve and Tcl_Release.
 *--------------------------------------------------------------
 */
static void PickCurrentItem(CanvasPlus *canvas, XEvent *event)
{
  int x, y, buttonDown;

  /*
   * Check whether or not a button is down.  If so, we'll log entry
   * and exit into and out of the current item, but not entry into
   * any other item.  This implements a form of grabbing equivalent
   * to what the X server does for windows.
   */
  buttonDown = Button1Mask|Button2Mask|Button3Mask|Button4Mask|Button5Mask;
  buttonDown &= canvas->state;
  if (!buttonDown) canvas->flags &= ~LEFT_GRABBED_ITEM;
  /*
   * Save information about this event in the canvas.  The event in
   * the canvas is used for two purposes:
   */
  if (event != &canvas->pickEvent) {
    if ((event->type == MotionNotify) || (event->type == ButtonRelease)) {
      canvas->pickEvent.xcrossing.type = EnterNotify;
      canvas->pickEvent.xcrossing.serial = event->xmotion.serial;
      canvas->pickEvent.xcrossing.send_event = event->xmotion.send_event;
      canvas->pickEvent.xcrossing.display = event->xmotion.display;
      canvas->pickEvent.xcrossing.window = event->xmotion.window;
      canvas->pickEvent.xcrossing.root = event->xmotion.root;
      canvas->pickEvent.xcrossing.subwindow = None;
      canvas->pickEvent.xcrossing.time = event->xmotion.time;
      canvas->pickEvent.xcrossing.x = event->xmotion.x;
      canvas->pickEvent.xcrossing.y = event->xmotion.y;
      canvas->pickEvent.xcrossing.x_root = event->xmotion.x_root;
      canvas->pickEvent.xcrossing.y_root = event->xmotion.y_root;
      canvas->pickEvent.xcrossing.mode = NotifyNormal;
      canvas->pickEvent.xcrossing.detail = NotifyNonlinear;
      canvas->pickEvent.xcrossing.same_screen = event->xmotion.same_screen;
      canvas->pickEvent.xcrossing.focus = False;
      canvas->pickEvent.xcrossing.state = event->xmotion.state;
    } else  {
      canvas->pickEvent = *event;
    }
  }
  /* If this is a reentrant call (there's already a partially completed
   * call pending on the stack;  it's in the middle of processing a
   * Leave event handler for the old current item) then just return;
   * the pending call will do everything that's needed.
   */
  if (canvas->flags & REPICK_IN_PROGRESS) return;
  /* A LeaveNotify event automatically means that there's no current
   * object, so the check for closest item can be skipped.
   */
  x = canvas->pickEvent.xcrossing.x;
  y = canvas->pickEvent.xcrossing.y;

  if (canvas->pickEvent.type != LeaveNotify)
    canvas->newCurrent = CPClosestItem(canvas, x, y);
  else
    canvas->newCurrent = NULL;

  /* Nothing to do:  the current item hasn't changed. */  
  if ((canvas->newCurrent==canvas->currentItem) &&
      !(canvas->flags & LEFT_GRABBED_ITEM))
    return;
  /*
   * Simulate a LeaveNotify event on the previous current item and
   * an EnterNotify event on the new current item.  Remove the "current"
   * tag from the previous current item and place it on the new current
   * item.
   */
  if ((canvas->newCurrent != canvas->currentItem) && 
      (canvas->currentItem != NULL) && !(canvas->flags & LEFT_GRABBED_ITEM)) {
    XEvent leave;
    CP_Item *item = canvas->currentItem;
    //    printf("Simulate leave (%p)\n", canvas->currentItem);

    leave = canvas->pickEvent;
    leave.type = LeaveNotify;
    leave.xcrossing.detail = NotifyAncestor;
    canvas->flags |= REPICK_IN_PROGRESS;
    ProcessEvent(canvas, &leave);
    canvas->flags &= ~REPICK_IN_PROGRESS;
    /* The check below is needed because there could be an event
     * handler for <LeaveNotify> that deletes the current item. */
    if ((item == canvas->currentItem) && !buttonDown)
      CPRemoveBindTag(item, CPGetCurrentTagUid());
  }
  if ((canvas->newCurrent != canvas->currentItem) && buttonDown) {
    canvas->flags |= LEFT_GRABBED_ITEM;
    return;
  }

  canvas->flags &= ~LEFT_GRABBED_ITEM;
  canvas->currentItem = canvas->newCurrent;

  if (canvas->currentItem != NULL)
    {
      XEvent enter;
      //      printf("Simulate enter (%p)\n", canvas->currentItem);
      CPAddBindTag(canvas->currentItem, CPGetCurrentTagUid());
      enter = canvas->pickEvent;
      enter.type = EnterNotify;
      enter.xcrossing.detail = NotifyAncestor;
      ProcessEvent(canvas, &enter);
    }
}


/*--------------------------------------------------------------
 * ProcessEvent --
 *	This procedure is called to invoke binding processing
 *	for a new event that is associated with the current item
 *	for a canvas.
 * Results:
 *	None.
 * Side effects:
 *	Depends on the bindings for the canvas.  A binding script
 *	could delete the canvas, so callers should protect themselves
 *	with Tcl_Preserve and Tcl_Release.
 *--------------------------------------------------------------
 */
#define NUM_STATIC 5

static void ProcessEvent(CanvasPlus *canvas, XEvent *event)
{
  ClientData staticObjects[NUM_STATIC];
  ClientData *obj;
  int nObj, i;
  CP_Item *item;

  if (canvas->mode == NULL) return;

  //  if ((event->type == KeyPress) || (event->type == KeyRelease)) {
  //  printf("KEY EVENT\n");
  //}
  //    item = canvas->textInfo.focusItem;
  //  else
    item = canvas->currentItem;

  if (item == NULL)  return;

  // Export some event information to the mode interpreter
  CPExportEventInfo(canvas, event);
  
  nObj = item->tags.number + 2;

  if (nObj <= NUM_STATIC)
    obj = staticObjects;
  else
    obj = (ClientData *)ckalloc((unsigned)(nObj * sizeof(ClientData)));

  obj[0] = (ClientData) CPGetAllTagUid();        // Tag for "all" must be always present

  for (i = item->tags.number-1; i >= 0; i--)     // pick all tags uid for this item
    obj[i+1] = (ClientData) item->tags.array[i];

  obj[item->tags.number+1] = (ClientData)item;  // include item ptr in the objects list
  
  if (canvas->tkwin != NULL) {
    Tk_BindEvent(canvas->mode->bindingTable, event, canvas->tkwin, nObj, obj);
  }
  if (obj != staticObjects) ckfree((char *) obj);
}

