/*
 * ***********************************************************************
 *
 * File 'tkCanvasPlusDrawing.c'.
 *
 * CanvasPlus widget drawing interface.
 *
 * Written by Jorge Lima in Nov 23 14:44:03 WET 2004
 *
 */

#include <X11/Xutil.h>
#include <tcl.h>
#include <tk.h>
#include <math.h>

#include "tkCanvasPlusDrawing.h"
#include "tkCanvasPlusTags.h"
#include "tkCanvasPlusModes.h"
#include "tkCanvasPlus.h"
#include "mathutils.h"


/***********************************************************************
 *
 * Rotines to manipulate the CanvasPlus item list
 *
 ***********************************************************************/

static void AppendItem(CP_Mode *mode, CP_Item *item)
{
  int isNew;
  CP_Item *tmp;
  Tcl_HashEntry *entry;
  item->id = mode->nextId;
  mode->nextId++;
  entry = Tcl_CreateHashEntry(&mode->idTable, (char *)item->id, &isNew);
  Tcl_SetHashValue(entry, item);

  tmp = mode->lastItem;
  mode->lastItem = item;
  item->prev = tmp;
  if ( item->prev == NULL ) // this is the first item
    mode->firstItem = item;
  else  // this is not the first item
    item->prev->next = item;
}

static void RemoveItem(CP_Mode *mode, CP_Item *item)
{
  Tcl_HashEntry *entry;
  entry = Tcl_FindHashEntry(&mode->idTable,(char *) item->id);
  Tcl_DeleteHashEntry(entry);

  if ( item->prev == NULL ) // this is the first item
    mode->firstItem = item->next;
  else                      // this is not the first item
    item->prev->next = item->next;

  if ( item->next == NULL ) // this is the last item
    mode->lastItem = item->prev;
  else                      // this is not the first item
    item->next->prev = item->prev;
}


/*------------------------------------------------------------------------
 *
 * Converting from physical to screen and from screen to physical
 *
 *------------------------------------------------------------------------
 */

inline static double windowToPhysicalX(CanvasPlus *cp, int x)
{ return (double)(x + cp->xOrigin) / (cp->x_scale * cp->zoom); }

inline static double windowToPhysicalY(CanvasPlus *cp, int y)
{ return (double)(y + cp->yOrigin) / (cp->y_scale * cp->zoom); }

inline static void windowToPhysical(CanvasPlus *cp, CP_Point *pp, XPoint *wp)
{
  pp->x = (double)(wp->x + cp->xOrigin) / (cp->x_scale * cp->zoom);
  pp->y = (double)(wp->y + cp->yOrigin) / (cp->y_scale * cp->zoom);
}

inline static int physicalToWindowX(CanvasPlus *cp, double x)
{ return (int)rint(x * cp->x_scale * cp->zoom) - cp->drawableXOrigin; }

inline static int physicalToWindowY(CanvasPlus *cp, double y)
{ return (int)rint(y * cp->y_scale * cp->zoom) - cp->drawableYOrigin; }

inline static void physicalToWindow(CanvasPlus *cp, XPoint *wp, CP_Point *pp)
{
  wp->x = (int)rint(pp->x * cp->x_scale * cp->zoom) - cp->drawableXOrigin;
  wp->y = (int)rint(pp->y * cp->y_scale * cp->zoom) - cp->drawableYOrigin;
}

inline static int physicalToWindowW(CanvasPlus *cp, double width)
{ return (int)ceil(cp->l_scale * cp->zoom * width); }

inline static double windowToPhysicalDist(CanvasPlus *cp, int dist)
{ return (double)dist/(cp->l_scale * cp->zoom); }
 
/*----------------------------------------------------------------------
 * CPInitPenSet --
 *	Initializes the Pen Set structure. Must be called before use
 *      of this structure.
 * Results:
 *	None.
 * Side effects:
 *	Pen Set structure will be left in a consistent state
 *---------------------------------------------------------------------*/
void CPInitPenSet(CP_Mode *mode)
{
  CP_PenSet *p = &mode->pens;
  p->penArray = p->penStatic;
  p->currentPen = NULL;
  p->penCount = 0;
  p->penArraySize = CP_STATIC_PEN_SET;
}

/*----------------------------------------------------------------------
 * CPCreatePen --
 *	Creates new Pen
 * Results:
 *	None.
 * Side effects:
 *---------------------------------------------------------------------*/
int CPCreatePen(CP_Mode *mode,  XGCValues *gcValues, unsigned long mask)
{
  CP_PenSet *ps = &mode->pens;
  Tk_Window tkwin = mode->tkwin;
  CP_Pen *newPen;

  if (ps->penCount == ps->penArraySize) {
    int newsize = ps->penArraySize+10;
    CP_Pen *penPtr;
    penPtr = (CP_Pen *)ckalloc((unsigned)(newsize*sizeof(CP_Pen)));
    if (penPtr==NULL) return CP_ALLOC_ERROR;
    memcpy((void *)penPtr, (void *)ps->penArray,
	   (unsigned)(ps->penCount*sizeof(CP_Pen)));
    if (ps->penArray != ps->penStatic)
      ckfree((char *)ps->penArray);
    ps->penArray = penPtr;
    ps->penArraySize = newsize;
  }

  newPen = &ps->penArray[ps->penCount];
  ps->penCount++;

  newPen->gc = XCreateGC(Tk_Display(tkwin),Tk_WindowId(tkwin), mask, gcValues);
  newPen->fcolor = NULL;
  newPen->bcolor = NULL;
  newPen->font = NULL;
  // return the index of the new element
  return ps->penCount-1;
}

/*----------------------------------------------------------------------
 * CPDeletePenSet --
 *	Deletes the structure 
 * Results:
 *	None.
 * Side effects:
 *	Eventually the memory is released
 *---------------------------------------------------------------------*/
void CPDeletePenSet(CP_Mode *mode)
{
  int i;
  CP_PenSet *p = &mode->pens;
  Tk_Window tkwin = mode->tkwin;
  
  for(i = 0; i < p->penCount; i++) {
    if (p->penArray[i].gc != None)
      XFreeGC(Tk_Display(tkwin), p->penArray[i].gc);
    if (p->penArray[i].fcolor != NULL)
      Tk_FreeColor(p->penArray[i].fcolor);
    if (p->penArray[i].bcolor != NULL)
      Tk_FreeColor(p->penArray[i].bcolor);
  }
  if (p->penArray != p->penStatic)
    ckfree((void *)p->penArray);
  p->penArray = p->penStatic;
  p->currentPen = NULL;
  p->penCount = 0;
  p->penArraySize = CP_STATIC_PEN_SET;
}

/*----------------------------------------------------------------------
 * CPSetUnit --
 *	This procedure defines the default working unit
 * Results:
 *	None.
 * Side effects:
 *	The CanvasPlus scale factors are changed
 *---------------------------------------------------------------------*/
void CPSetUnit(CanvasPlus *cp, double unit)
{
  Tk_Window tkwin = cp->tkwin;
  cp->x_scale = unit * DisplayWidth(Tk_Display(tkwin), Tk_ScreenNumber(tkwin))/
    DisplayWidthMM(Tk_Display(tkwin), Tk_ScreenNumber(tkwin)); 
  cp->y_scale = unit * DisplayHeight(Tk_Display(tkwin), Tk_ScreenNumber(tkwin))/
    DisplayHeightMM(Tk_Display(tkwin), Tk_ScreenNumber(tkwin));
  /* for line scale use the smallest */
  cp->l_scale = dmin(cp->x_scale, fabs(cp->y_scale));
  /* make sure the cp->y_scale has the correct sign */
  if (cp->yInvert)
    cp->y_scale = -fabs(cp->y_scale);
  else
    cp->y_scale = fabs(cp->y_scale);
}

/*----------------------------------------------------------------------
 * CPRequestCanvasDimensions --
 *	This procedure adjusts the physical canvas+ dimensions to the
 *      specified to the physical ones. Must be called after a zoom.
 * Results:
 *	None.
 * Side effects:
 *	The CanvasPlus area in screen dimensions will be updated.
 *---------------------------------------------------------------------*/
void CPRecomputeCanvasDimensions(CanvasPlus *cPtr)
{
  double xscale = cPtr->x_scale * cPtr->zoom;
  double yscale = cPtr->y_scale * cPtr->zoom;
  cPtr->scrollX1 = (int)rint(xscale * cPtr->Xmin);
  cPtr->scrollY1 = (int)rint(yscale * cPtr->Ymin);
  cPtr->scrollX2 = (int)rint(xscale * cPtr->Xmax);
  cPtr->scrollY2 = (int)rint(yscale * cPtr->Ymax);
  CPForceRedraw(cPtr);
}

/*----------------------------------------------------------------------
 * CPRequestPhysicalDimensions --
 *	This procedure adjusts the physical canvas+ dimensions to
 *      accomodate the specified physical world limits
 * Results:
 *	None.
 * Side effects:
 *	The CanvasPlus area in screen dimensions will be updated.
 *---------------------------------------------------------------------*/
void CPRequestPhysicalDimensions(CanvasPlus *cPtr, const CP_Area *area)
{
  cPtr->Xmin = area->x1;
  cPtr->Ymin = area->y1;
  cPtr->Xmax = area->x2;
  cPtr->Ymax = area->y2;
  CPRecomputeCanvasDimensions(cPtr);
}

/*----------------------------------------------------------------------
 * CPExpandPhysicalDimensions --
 *	This procedure expands the physical canvas+ dimensions to
 *      to include the box specified
 * Results:
 *	None.
 * Side effects:
 *	The CanvasPlus area in screen dimensions will be updated.
 *---------------------------------------------------------------------*/
void CPExpandPhysicalDimensions(CanvasPlus *cPtr, const CP_Area *area)
{
  if (cPtr->Xmin > area->x1) cPtr->Xmin = area->x1;
  if (cPtr->Ymin > area->y1) cPtr->Ymin = area->y1;
  if (cPtr->Xmax < area->x2) cPtr->Xmax = area->x2;
  if (cPtr->Ymax < area->y2) cPtr->Ymax = area->y2;
  CPRecomputeCanvasDimensions(cPtr);
}

void CPExportEventInfo(CanvasPlus *cp, XEvent *event)
{
  int i;
  CP_Mode *mode = cp->mode;
  double x = windowToPhysicalX(cp, event->xmotion.x);
  double y = windowToPhysicalY(cp, event->xmotion.y);
  Tcl_SetVar2Ex(mode->interp, "_X_", NULL, Tcl_NewDoubleObj(x), TCL_GLOBAL_ONLY);
  Tcl_SetVar2Ex(mode->interp, "_Y_", NULL, Tcl_NewDoubleObj(y), TCL_GLOBAL_ONLY);
  if (cp->currentItem != NULL) {
    i = cp->currentItem->id;
    Tcl_SetVar2Ex(mode->interp, "_I_", NULL, Tcl_NewIntObj(i), TCL_GLOBAL_ONLY);
  }
  else {
    Tcl_UnsetVar(mode->interp, "_I_", TCL_GLOBAL_ONLY);
  }
}

/*----------------------------------------------------------------------
 * CPCreateItem --
 *	This function is called to create a new CanvasPlus Item. 
 * Results:
 *	Pointer the CP_Item or NULL
 * Side effects:
 *	The CP_Item structure is created in the CanvasPlus that will be
 *      used for canvas bindings. But its BBOX is undefined. must be set
 *      and CPUpdateItem must be called when finished with initialization
 *---------------------------------------------------------------------*/
CP_Item *CPCreateItem(CanvasPlus *cp, size_t size)
{
  CP_Mode *mode = cp->mode;
  CP_Item *item;

  if (mode == NULL)
    return NULL;

  if (size < 0) return NULL;
  item = (CP_Item *)ckalloc(sizeof(CP_Item)+size);
  if (item!=NULL) {
    item->clientData = NULL;
    item->next = NULL;
    item->prev = NULL;
    CPInitItemTags(item);
  }
  if (size != 0)  // the private data is allocad here, 
                  // so initialize item->clientData
    item->clientData = (ClientData)(item+1);

  AppendItem(mode,item);
  return item;
}
 
/**************************************************************************
 *
 * Must be called when the item is deleted to release its related 
 * information in the canvas.
 *
 **************************************************************************/
void CPDeleteItem(CanvasPlus *canvas, CP_Item *cpi)
{
  CP_Mode *mode = canvas->mode;
  if (mode==NULL) return;
  RemoveItem(mode, cpi);
  CPDeleteItemTags(cpi);
  ckfree((void *)cpi);
}

void CPDeleteAllItems(CanvasPlus *canvas)
{
  CP_Item *prev;
  CP_Mode *mode = canvas->mode;
  if (mode != NULL) {
    CP_Item *item = mode->firstItem;
    while ( item != NULL ) {
      prev = item;
      item = item->next;
      CPDeleteItemTags(prev);
      RemoveItem(mode, prev);
      ckfree((void *)prev);
    }
    mode->firstItem = NULL;
    mode->lastItem = NULL;
    mode->nextId = 0;
  }
}

Tcl_Obj *CPListAllItems(CP_Mode *mode)
{
  CP_Item *item = mode->firstItem;
  Tcl_Obj *obj = Tcl_NewObj();
  while (item != NULL) {
    Tcl_ListObjAppendElement(NULL, obj, Tcl_NewIntObj(item->id));
    item = item->next;
  }
  return obj;
}

Tcl_Obj *CPListTaggedItems(CP_Mode *mode, Tk_Uid tagUid)
{
  CP_Item *item = mode->firstItem;
  Tcl_Obj *obj = Tcl_NewObj();
  while (item != NULL) {
    item = CPSearchBindTag(item, tagUid);
    if (item == NULL) break;
    Tcl_ListObjAppendElement(NULL, obj, Tcl_NewIntObj(item->id));
    item = item->next;
  }
  return obj;
}

/*----------------------------------------------------------------------
 * CPGetItemFromId --
 *      Gets the item corresponding to the given ID
 * Results:
 *	Pointer to CP_Item or NULL.
 * Side effects:
 *	None
 *---------------------------------------------------------------------*/
CP_Item *CPGetItemFromId(CP_Mode *mode, int id)
{
  CP_Item *item = NULL;
  Tcl_HashEntry *entry;
  entry = Tcl_FindHashEntry(&mode->idTable, (char *)id);
  if (entry != NULL)
    item = (CP_Item *) Tcl_GetHashValue(entry);
  return item;
}

/*********************************************************************
 *
 * Draw rotines
 *
 *********************************************************************/

static int CPSelectPen(ClientData d, int index)
{
  CanvasPlus *canvas = (CanvasPlus *)d;
  CP_PenSet *set = &canvas->mode->pens;
  if (index < set->penCount) {
    set->currentPen = &set->penArray[index];
    return CP_OK;
  } else {
    return CP_PEN_ERROR;
  }
}

static int CPSetPenColor(ClientData d, char *foreground, char *background)
{
  CanvasPlus *cp = (CanvasPlus *)d;
  Tk_Window tkwin = cp->tkwin;
  CP_Mode *mode = cp->mode;
  Tcl_Interp *interp = mode->interp;
  CP_Pen *pen = mode->pens.currentPen;
  XColor *xcolor;

  if (pen == NULL)
    return CP_PEN_ERROR;
  
  xcolor = Tk_GetColor(interp, tkwin, foreground);

  if (xcolor==NULL) return CP_COLOR_ERROR;

  if (pen->fcolor == NULL) {
    pen->fcolor = xcolor;
    XSetForeground(Tk_Display(cp->tkwin), pen->gc, pen->fcolor->pixel);
  }
  else if (xcolor->pixel == pen->fcolor->pixel) {
    Tk_FreeColor(xcolor);
  }
  else {
    Tk_FreeColor(pen->fcolor);
    pen->fcolor = xcolor;
    XSetForeground(Tk_Display(cp->tkwin), pen->gc, pen->fcolor->pixel);
  }

  xcolor = Tk_GetColor(interp, tkwin, background);

  if (xcolor==NULL)
    return CP_COLOR_ERROR;

  if (pen->bcolor == NULL) {
    pen->bcolor = xcolor;
    XSetBackground(Tk_Display(cp->tkwin), pen->gc, pen->bcolor->pixel);
  }
  else if (xcolor->pixel == pen->bcolor->pixel) {
    Tk_FreeColor(xcolor);
  }
  else {
    Tk_FreeColor(pen->bcolor);
    pen->bcolor = xcolor;
    XSetBackground(Tk_Display(cp->tkwin), pen->gc, pen->bcolor->pixel);
  }
  return CP_OK;
}

static void CPSetPenWidth(ClientData d, double width)
{
  CanvasPlus *canvas = (CanvasPlus *)d;
  CP_Mode *mode = canvas->mode;
  CP_Pen *pen = mode->pens.currentPen;
  XGCValues values;
  
  if (pen != NULL) {
    if (width == 0.0)
      values.line_width = 1;
    else
      values.line_width = physicalToWindowW(canvas, width);
 
    XChangeGC(Tk_Display(canvas->tkwin), pen->gc, GCLineWidth, &values);
    //    XSetLineAttributes(Tk_Display(canvas->tkwin), pen->gc, w,
    //		       LineSolid, CapRound, JoinRound);
  }
}

static void CPDrawPolyline(ClientData d, int npoints, CP_Point *ppts)
{
  CanvasPlus *canvas = (CanvasPlus *)d;
  CP_Mode *mode = canvas->mode;
  CP_Pen *pen = mode->pens.currentPen;

  if (pen != NULL) {
    int n;
    XPoint points[npoints];
    Display *display = Tk_Display(canvas->tkwin);
    for (n = 0; n < npoints; n++)
      physicalToWindow(canvas, &points[n], &ppts[n]);

    XDrawLines(display, canvas->drawable, pen->gc,
	       points, npoints, CoordModeOrigin);
  }
}

static void CPDrawPoints(ClientData d, int npoints, CP_Point *ppts)
{
  CanvasPlus *canvas = (CanvasPlus *)d;
  CP_Mode *mode = canvas->mode;
  CP_Pen *pen = mode->pens.currentPen;

  if (pen != NULL) {
    int n;
    XPoint points[npoints];
    Display *display = Tk_Display(canvas->tkwin);
    for (n = 0; n < npoints; n++)
      physicalToWindow(canvas, &points[n], &ppts[n]);

    XDrawPoints(display, canvas->drawable, pen->gc,
		points, npoints, CoordModeOrigin);
  }
}

static void CPDrawPolygon(ClientData d, int npoints, CP_Point *ppts,
			  int shape, int draw_mode)
{
  CanvasPlus *canvas = (CanvasPlus *)d;
  CP_Mode *mode = canvas->mode;
  CP_Pen *pen = mode->pens.currentPen;

  if (pen != NULL) {
    int n;
    XPoint points[npoints+1];
    Display *display = Tk_Display(canvas->tkwin);
    for (n = 0; n < npoints; n++)
      physicalToWindow(canvas, &points[n], &ppts[n]);

    physicalToWindow(canvas, &points[npoints], &ppts[0]);
    if (draw_mode == CP_DRAW_FILLED)
      XFillPolygon(display, canvas->drawable, pen->gc,
		   points, npoints, shape, CoordModeOrigin);
    else // CP_DRAW_NO_OUTLINE
      XDrawLines(display, canvas->drawable, pen->gc,
		 points, npoints+1, CoordModeOrigin);
  }
}

static void CPDrawRectangle(ClientData d, CP_Area *rect, int draw_mode)
{
  CanvasPlus *canvas = (CanvasPlus *)d;
  CP_Mode *mode = canvas->mode;
  CP_Pen *pen = mode->pens.currentPen;

  if (pen != NULL) {
    int x, y, width, height;
    Display *display = Tk_Display(canvas->tkwin);
    x = physicalToWindowX(canvas, rect->x1);
    y = physicalToWindowY(canvas, rect->y1);
    width = physicalToWindowX(canvas, rect->x2) - x;
    height = physicalToWindowY(canvas, rect->y2) - y;
    if (width < 1 || height < 1) return;
    if (draw_mode == CP_DRAW_FILLED)
      XFillRectangle(display, canvas->drawable, pen->gc,
		     x, y, (unsigned)width, (unsigned)height);
    else // CP_DRAW_OUTLINE
      XDrawRectangle(display, canvas->drawable, pen->gc,
		     x, y, (unsigned)width, (unsigned)height);
  }
}

static void CPDrawElipse(ClientData d, CP_Area *rect, int draw_mode)
{
  CanvasPlus *canvas = (CanvasPlus *)d;
  CP_Mode *mode = canvas->mode;
  CP_Pen *pen = mode->pens.currentPen;

  if (pen != NULL) {
    int x, y, width, height;
    Display *display = Tk_Display(canvas->tkwin);
    x = physicalToWindowX(canvas, rect->x1);
    y = physicalToWindowY(canvas, rect->y1);
    width = physicalToWindowX(canvas, rect->x2) - x;
    height = physicalToWindowY(canvas, rect->y2) - y;
    if (width < 1 || height < 1) return;
    if (draw_mode == CP_DRAW_FILLED)
      XFillArc(display, canvas->drawable, pen->gc, x, y,
	       (unsigned)width, (unsigned)height, 0, 360 * 64);
    else // CP_DRAW_OUTLINE
      XDrawArc(display, canvas->drawable, pen->gc, x, y,
	       (unsigned)width, (unsigned)height, 0, 360 * 64);
  }
}

/*
void CPSetFont(ClientData d, char *fontname)
{
   CanvasPlus *canvas = (CanvasPlus *)d;
   XGCValues values;
   CP_Mode *mode = canvas->mode;
   CP_Pen *pen = mode->pens.currentPen;
   if (pen->font != 0)
     Tk_FreeFont(pen->font);
   pen->font = Tk_GetFont(interp, canvas->tkwin, fontname);
   values.font = Tk_FontId(pen->font);
   XChangeGC(Tk_Display(canvas->tkwin), pen->gc, GCFont, &values);
}
*/

void CPDrawText(ClientData d, double x, double y, char *text)
{
  int wx, wy;
  CanvasPlus *canvas = (CanvasPlus *)d;
  XGCValues values;
  CP_Mode *mode = canvas->mode;
  CP_Pen *pen = mode->pens.currentPen;
  wx = physicalToWindowX(canvas, x);
  wy = physicalToWindowY(canvas, y);
  XDrawString(Tk_Display(canvas->tkwin), canvas->drawable,
	      pen->gc, wx, wy, text, strlen(text));
}


static CP_DrawOps _CanvasDrawOps = {
  NULL,
  CPSelectPen,
  CPSetPenColor,
  CPSetPenWidth,
  NULL, /* setFont() */
  NULL, /* setFontSize() */
  CPDrawPoints,
  CPDrawPolyline,
  CPDrawPolygon,
  CPDrawRectangle,
  CPDrawElipse,
  NULL  /* drawText() */
};

/*----------------------------------------------------------------------
 * CPDrawContents --
 *	This procedure will draw the canvas+ contents
 * Results:
 *	None.
 * Side effects:
 *	Everything in the visible region of canvas+ will be redrawn.
 *---------------------------------------------------------------------*/
void CPDrawContents(CanvasPlus *canvas)
{
  CP_Area redraw;
  CP_Mode *mode = canvas->mode;
  if (mode != NULL) {
    CP_ModeOps *ops = mode->ops;
    ClientData clientData = mode->clientData;
    CP_DrawOps *drawOps = &_CanvasDrawOps;

    redraw.x1 = windowToPhysicalX(canvas, canvas->redrawX1-canvas->drawableMargin);
    redraw.x2 = windowToPhysicalX(canvas, canvas->redrawX2+canvas->drawableMargin);
    redraw.y1 = windowToPhysicalY(canvas, canvas->redrawY1-canvas->drawableMargin);
    redraw.y2 = windowToPhysicalY(canvas, canvas->redrawY2+canvas->drawableMargin);
    
    drawOps->clientData = (ClientData)canvas;  
    ops->drawProc(clientData, drawOps, &redraw);
  }
}

/*----------------------------------------------------------------------
 * CPClosestItem --
 *	Given x and y coordinates, find the topmost canvas item that
 *	is "close" to the coordinates. Canvas items that are hidden
 *	or disabled are ignored.
 * Results:
 *	The return value is a pointer to the topmost item that is
 *	close to (x,y), or NULL if no item is close.
 * Side effects:
 *	None.
 *----------------------------------------------------------------------
 */
CP_Item *CPClosestItem(CanvasPlus *canvas, int x, int y)
{
  CP_Mode *mode = canvas->mode;
  if (mode != NULL) {/* mode haven't been initialized yet */
    CP_Point pt;
    CP_ModeOps *ops = mode->ops;
    ClientData clientData = mode->clientData;
    double dist = windowToPhysicalDist(canvas, canvas->closeEnough);
    pt.x = windowToPhysicalX(canvas, x);
    pt.y = windowToPhysicalY(canvas, y);
    return ops->closestProc(clientData, &pt, dist);
  }
  return NULL;
}

