package require smtp
package require mime

set smtpHost zodiac.maf.nasa.gov
set emailFromLine {License Manager}
set maxState 0
set currentState 0
set deadbandActive 0
set deadbandTimerId {}
set lastSave [clock seconds]
set logFileName {}
set dataDir [file normalize [file join [pwd] etc]]
set dataFile [file join $dataDir LicenseManager.acd]
##
## Arguments to logMsgFormat are:
##      1 - time in seconds sicne epoch
##      2 - human readable time
##      3 - message
##
set logMsgFormat {[%1$s] -- %2$s -- %3$s.}
##
## Arguments to logAlarmFormat, emailBodyFormat and emailSubjectFormat are:
##      1 - time in seconds sicne epoch
##      2 - human readable time
##      3 - license count
##      4 - old state name
##      5 - old state description
##      6 - new state name
##      7 - new state description
##
set logAlarmFormat {[%1$s] -- %2$s -- Entering %$6s from %4$s. License count is %3$s.}
set emailBodyFormat $logAlarmFormat
set emailSubjectFormat {License Manager Alert -- Entering %$6s from %4$s}

##
## Valid alarm indexes are:
##
##     <id>,alarmName
##     <id>,alarmDescription
##     <id>,licenseCount
##     <id>,up,emailToList
##     <id>,down,emailToList
##     <id>,deadbandTime
##     <id>,deadbandLimit
##     <id>,fgColor
##     <id>,bgColor
##
## where ids range from 0 up
##
array set alarmDataArray {
    0,alarmName         {Normal}
    0,alarmDescription  {No Alarms Active}
    0,licenseCount      0
    0,up,emailToList    {}
    0,down,emailToList   {}
    0,deadbandTime      {}
    0,deadbandLimit     -1
    0,fgColor           Green
    0,bgColor           White
}


###########################################################################
#
# Procedure Name : AddAlarm
#
# Description : Adds a new alarm to the system.
#
# Arguments :
#     alarmName        -- Short name of alarm
#     alarmDescription -- Description of alarm
#     licenseCount     -- Limit at which this alarm is active
#     up_emailToList   -- List of people to send and email to when the
#                         alarm is activated riseing into from an alarm
#                         with a lower limit
#     down_emailToList -- List of people to send and email to when the
#                         alarm is activated by dropping into from a
#                         alarm with a higher limit
#     deadbandTime     -- How many seconds much pass before the alarm
#                         level is allowed to drop to a lower level
#     deadbandLimit    -- What value the license count must drop below to
#                         allow the alarm level is allowed to drop to a
#                         lower level
#     fgColor          -- The foreground color of the alarm
#     bgColor          -- The background color of the alarm
#
# Returns : nonthing
#
# Side-Effects : Writes out alarm settings
#
# Exception Conditions : none
#
# Pre-requisite Conditions :
#
# Original Author : Gerald W. Lester
#
###########################################################################
proc AddAlarm {alarmName alarmDescription licenseCount up_emailToList down_emailToList deadbandTime deadbandLimit fgColor bgColor} {
    variable maxState
    variable alarmDataArray
    variable currentState
    upvar #0 totalConnectionCount currentLicenseCount

    ##
    ## Build current alarms as list of list
    ##
    set stateList [list]
    for {set i 0} {$i <= $maxState} {incr i 1} {
        set alarm [list $i]
        foreach key {alarmName alarmDescription licenseCount up,emailToList down,emailToList deadbandTime deadbandLimit fgColor bgColor} {
            lappend alarm $alarmDataArray($i,$key)
        }
        lappend stateList $alarm
    }
    puts "State List = {$stateList}"

    ##
    ## Add new alarm, mark alarmId as @
    ##
    set alarm [list {@}]
    foreach key {alarmName alarmDescription licenseCount up_emailToList down_emailToList deadbandTime deadbandLimit fgColor bgColor} {
        lappend alarm [set $key]
    }
    lappend stateList $alarm
    puts "State List = {$stateList}"

    ##
    ## Sort them by licenseCount
    ##
    set stateList [lsort -index 3 -integer $stateList]
    puts "State List = {$stateList}"

    ##
    ## Store the new order
    ##
    set newAlarmId -1
    set i -1
    foreach alarm $stateList {
        puts "Processing {$alarm}"
        incr i
        set key [lindex $alarm 0]
        set alarm [lrange $alarm 1 end]
        if {[string equal $key {@}]} {
            set newAlarmId $i
            set key $i
        }
        foreach element {alarmName alarmDescription licenseCount up,emailToList down,emailToList deadbandTime deadbandLimit fgColor bgColor} value $alarm {
            set alarmDataArray($i,$element) $value
        }
    }
    parray alarmDataArray

    ##
    ## Bump the current level if needed be
    ##
    if {$newAlarmId <= $currentState} {
        incr currentState
    } elseif {($newAlarmId == $currentState) &&
              ($alarmDataArray([expr {$newAlarmId + 1}],licenseCount) >= $currentLicenseCount)} {
        incr currentState
    }
    incr maxState
    WriteAlarmSettings
    return;
}


###########################################################################
#
# Procedure Name : CheckForChange
#
# Description : Checks to see if the database has changed during an edit.
#
# Arguments :
#     checkId -- Timestamp of last write at time of page generation.
#
# Returns : nonthing if all ok
#
# Side-Effects : Writes out alarm settings
#
# Exception Conditions : If timestamps do not match, raises an error
#
# Pre-requisite Conditions :
#
# Original Author : Gerald W. Lester
#
###########################################################################
proc CheckForChange {checkId} {
    variable lastSave

    if {$lastSave != $checkId} {
        ##
        ## Page was invalid, generate error HTML
        ##
        set html {
            <h2>Save Failed, alarm database was modified before delete -- <A HREF="Alarms.tml">Click Here to Continue</A></h2>
        }
        return -code error $html
    }
    return;
}


###########################################################################
#
# Procedure Name : DeleteAlarm
#
# Description : Delete an alarm.
#
# Arguments :
#     alarmId -- Id of alarm to delete.
#     checkId -- Timestamp of last write at time of page generation.
#
# Returns : HTML containing status of deletion
#
# Side-Effects : Writes out alarm settings
#
# Exception Conditions : none
#
# Pre-requisite Conditions :
#
# Original Author : Gerald W. Lester
#
###########################################################################
proc DeleteAlarm {alarmId checkId} {
    variable maxState
    variable alarmDataArray
    variable currentState

    set html {}
    if {![catch {CheckForChange $checkId} html]} {
        ##
        ## Drop the current level if needed be
        ##
        if {$currentState == $alarmId} {
            incr currentState -1
        }

        ##
        ## Roll the everything down one
        ##
        for {set i $alarmId; set j [expr {$i + 1}]} {$i < $maxState} {set i $j; incr j} {
            foreach element {alarmName alarmDescription licenseCount up,emailToList down,emailToList deadbandTime deadbandLimit fgColor bgColor} {
                set alarmDataArray($i,$element) $alarmDataArray($j,$element)
            }
        }

        ##
        ## Unset the last item
        ##
        array unset alarmDataArray $maxState,*
        incr maxState -1

        ##
        ## Generate Success HTML
        ##
        set html {
            <h2>Save Successful -- <A HREF="Alarms.tml">Click Here to Continue</A></h2>
        }
        WriteAlarmSettings
    }
    return $html
}


###########################################################################
#
# Procedure Name : CheckAlarms
#
# Description : Check to see if any alarms.
#
# Arguments :
#     args -- No used, ignored.
#
# Returns : nonthing
#
# Side-Effects : none
#
# Exception Conditions : none
#
# Pre-requisite Conditions :
#
# Original Author : Gerald W. Lester
#
###########################################################################
proc CheckAlarms {args} {
    variable maxState
    variable currentState
    variable deadbandTimerId
    variable deadbandActive
    variable alarmDataArray
    upvar #0 totalConnectionCount licenseCount

    ##
    ## Find the last state that the current license count is greater than
    ##
    for {set i $maxState} {$licenseCount < $alarmDataArray($i,licenseCount)} {incr i -1} {}
    set newState $i

    ##
    ## See if we are at an elevated alarm state or have dropped
    ##
    if {$newState < $currentState} {
        ##
        ## At Lowered Level
        ##
        if {$licenseCount < $alarmDataArray($currentState,deadbandLimit)} then {
            ##
            ## Clear Deadband
            ##
            after cancel $deadbandTimerId
            set deadbandTimerId {}
            set deadbandActive 0
        }
        if {!$deadbandActive} {
            ##
            ## Deadband not active, so allow drop of level
            ##
            WriteAlarmLog $currentState $newState $licenseCount
            SendAlarmEmail $newState down $licenseCount
            set currentState $newState
        }
    } else {
        ##
        ## At the same level or higher, so reset deadband timer
        ##
        after cancel $deadbandTimerId
        set deadbandTimerId {}
        set deadbandActive 0
        set time $alarmDataArray($newState,deadbandTime)
        if {$[string length $time]} {
            set deadbandTimerId [after [expr {$time * 1000}] {
                set deadbandTimerId {}
                set deadbandActive 0
            }
        }
        if {$newState > $currentState} {
            ##
            ## At Elevated Alarm
            ##
            WriteAlarmLog $currentState $newState $licenseCount
            SendAlarmEmail $newState up $licenseCount
            set currentState $newState
        }
    }
    set fgColor $alarmDataArray($currentState,fgColor)
    set bgColor $alarmDataArray($currentState,bgColor)

    return;
}


###########################################################################
#
# Procedure Name : SendAlarmEmail
#
# Description : Send an email.
#
# Arguments :
#     state -- New alarm level.
#     direction -- Direction of change: up or down.
#     licenseCount -- new number of licenses uset.
#
# Returns : nothing
#
# Side-Effects : none
#
# Exception Conditions : none
#
# Pre-requisite Conditions :
#
# Original Author : Gerald W. Lester
#
###########################################################################
proc SendAlarmEmail {state direction licenseCount} {
    variable alarmDataArray
    variable smtpHost
    variable emailFromLine
    variable emailBodyFormat
    variable emailSubjectFormat

    if {[string length $smtpHost] &&
        [string length $alarmDataArray($state,$drection,emailToList)]} {
        set toList $alarmDataArray($state,$drection,emailToList)
        set now [clock seconds]
        set body [format $emailBodyFormat \
                        $now \
                        [clock format $now] \
                        $licenseCount \
                        $alarmDataArray($oldState,alarmName) \
                        $alarmDataArray($oldState,alarmDescription) \
                        $alarmDataArray($newState,alarmName) \
                        $alarmDataArray($newState,alarmDescription) \
        ]
        set subject [format $emailSubjectFormat \
                        $now \
                        [clock format $now] \
                        $licenseCount \
                        $alarmDataArray($oldState,alarmName) \
                        $alarmDataArray($oldState,alarmDescription) \
                        $alarmDataArray($newState,alarmName) \
                        $alarmDataArray($newState,alarmDescription) \
        ]
        set mime [::mime::initialize -canonical text/plain -string $body]
        mime::setheader $mime Subject $subject
        catch {
            ::smtp::sendmessage  $mime \
                -servers $smtpHost \
                -originator $emailFromLine \
                -recipients $toList
        }
        ::mime::finalize $mime
    }

    return;
}


###########################################################################
#
# Procedure Name : WriteAlarmLog
#
# Description : Write an alarm message to the alarm log.
#
# Arguments :
#     oldState -- Old alarm level.
#     newState -- New alarm level.
#     licenseCount -- new number of licenses uset.
#
# Returns : nonthing
#
# Side-Effects : none
#
# Exception Conditions : none
#
# Pre-requisite Conditions :
#
# Original Author : Gerald W. Lester
#
###########################################################################
proc WriteAlarmLog {oldState newState licenseCount} {
    variable alarmDataArray
    variable logFileName
    variable logAlarmFormat

    if {[string length $logFileName]} {
        set ofd [open [subst -nobackslash $logFileName] {WRONLY APPEND CREAT}]
        set now [clock seconds]
        puts $ofd [format $logAlarmFormat \
                        $now \
                        [clock format $now] \
                        $licenseCount \
                        $alarmDataArray($oldState,alarmName) \
                        $alarmDataArray($oldState,alarmDescription) \
                        $alarmDataArray($newState,alarmName) \
                        $alarmDataArray($newState,alarmDescription) \
        ]
        close $ofd
    }
}


###########################################################################
#
# Procedure Name : WriteLog
#
# Description : Write a message to the alarm log.
#
# Arguments :
#     message -- message.
#
# Returns : nonthing
#
# Side-Effects : none
#
# Exception Conditions : none
#
# Pre-requisite Conditions :
#
# Original Author : Gerald W. Lester
#
###########################################################################
proc WriteLog {message} {
    variable alarmDataArray
    variable logFileName
    variable logMsgFormat

    if {[string length $logFileName]} {
        set ofd [open [subst -nobackslash $logFileName] {WRONLY APPEND CREAT}]
        set now [clock seconds]
        puts $ofd [format $logMsgFormat \
                        $now \
                        [clock format $now] \
                        $message
        ]
        close $ofd
    }
}


###########################################################################
#
# Procedure Name : WriteAlarmSettings
#
# Description : Writes the alarm settigns for later use..
#
# Arguments : none
#
# Returns : nonthing
#
# Side-Effects : none
#
# Exception Conditions : none
#
# Pre-requisite Conditions :
#
# Original Author : Gerald W. Lester
#
###########################################################################
proc WriteAlarmSettings {} {
    variable dataFile
    variable alarmDataArray
    variable lastSave
    variable maxState

    set lastSave [clock seconds]
    set ofd [open $dataFile {WRONLY TRUNC CREAT}]
    puts $ofd [list array set alarmDataArray [array get alarmDataArray]]
    puts $ofd [list set maxState $maxState]
    puts $ofd [list set lastSave $lastSave]
    close $ofd
}

